// BeOS driver for fake H/W Monitoring IC


#include <KernelExport.h>
#include <Drivers.h>
#include <Errors.h>
#include <stdio.h>

#define DEBUG 1

#include <Debug.h>

/*
#ifdef DEBUG
	#define DEBUG_ONLY(expr)	{ expr; }
#else
	#define DEBUG_ONLY(expr)	NULL
#endif
*/

#include "Sensors.h"

#include "fake_sensor.h"


#ifdef CODEWARRIOR
	#pragma mark [Local definitions]
#endif

// Local definitions
// -----------------


#ifndef DRIVER_NAME
	#define DRIVER_NAME 		"generic"
#endif
#define DEBUGGER_COMMAND	 	DRIVER_NAME

#define DEVICE_NAME 			DRIVER_NAME
#define DEVICE_NAME_LENGTH		(64)			
#define DEVICE_PATH 			"misc/sensor/" DEVICE_NAME "/"
#define DEVICE_MAX_ENTRIES 		(2)			

typedef struct device {
	long			id;
	const char *	name;
	uint32			flags;
	uint32			debug_level;

	// Locks mechanisms
	uint32			open_count;
	cpu_status		cs;					// Tempory storage for disable_interrupt()
	spinlock		hardware_spinlock;	// Hardware access *hardlock*
	sem_id			init_semaphore;		// Semaphore for init locking
	sem_id			update_semaphore;	// Semaphore for update locking
	bigtime_t		last_updated;		// Time we read values, 0 = never

	// Monitor datas
	uint32			nb_sensors;
	SENSOR *		sensors;
	void *			data;				// sensors chip specific datas	
	
	// Somes statistics
	volatile		nb_updates;
} device;

	
// globals for driver instances
static char *				g_device_name_list[DEVICE_MAX_ENTRIES + 1] 	= { NULL };
static device *				g_device_list[DEVICE_MAX_ENTRIES + 1] 		= { NULL };

static FAKE 	g_fake[] = { "faked hardware monitoring chip", 0x1234 };

// Prototypes of device hooks functions
static status_t open_hook(const char *name, uint32 flags, void ** cookie);
static status_t close_hook(void * cookie);
static status_t free_hook(void * cookie);
static status_t control_hook(void * cookie, uint32 msg,void * buf, size_t len);
static status_t read_hook(void * cookie, off_t pos, void * buf, size_t * len);
static status_t write_hook(void * cookie, off_t pos, const void * buf, size_t * len);

/* And, so, the device hooks list */
static device_hooks g_device_hooks = {
	open_hook, 			/* -> open entry point */
	close_hook,         /* -> close entry point */
	free_hook,          /* -> free entry point */
	control_hook,       /* -> control entry point */
	read_hook,          /* -> read entry point */
	write_hook,         /* -> write entry point */
	NULL,               /* -> select entry point */
	NULL,               /* -> deselect entry point */
	NULL,               /* -> readv */
	NULL                /* -> writev */
};

// privates prototypes
// -------------------

static status_t 	device_init(device * dev);
static status_t		device_uninit(device * dev);

static SENSOR *		device_find_sensor(device *	dev, char * name, int id);

static status_t 	fake_update_values(device * dev);
static status_t 	fake_init(device * dev);
static status_t 	fake_uninit(device * dev);
static int			fake_read(FAKE * f, int reg);
static status_t		fake_write(FAKE * f, int reg, int value);

#ifdef DEBUGGER_COMMAND
	static int 		device_debugger_command(int argc, char ** argv);
#endif

static status_t 	device_clear_statistics(device * dev);

#ifdef CODEWARRIOR
	#pragma mark [Driver API calls]
#endif

// Driver API calls
// ----------------


_EXPORT int32 api_version = B_CUR_DRIVER_API_VERSION;


// -------------------------------------------------------
_EXPORT status_t init_hardware (void)
{
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- init_hardware()\n") );

	return B_OK;
}



// -------------------------------------------------------
_EXPORT status_t init_driver(void)
{
	long		nb;
	char		name[ DEVICE_NAME_LENGTH + 1 ];
	device *	dev;

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- init_driver(), built %s %s\n", __DATE__, __TIME__) );

	for ( nb = 0; nb < DEVICE_MAX_ENTRIES; nb++ )
		{
		sprintf(name, "%s%ld", DEVICE_PATH, nb);

		g_device_name_list[nb] 	= (char *) malloc(strlen(name) + 1);
		g_device_list[nb] 		= (device *) malloc(sizeof(device));

		// Check that mallocs succeed...
		if ( (NULL == g_device_name_list[nb]) ||
			 (NULL == g_device_list[nb])	)
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": malloc() FAILED -> ABORT !") );

			if ( NULL != g_device_name_list[nb] )
				free(g_device_name_list[nb]);
			if ( NULL != g_device_list[nb] )
				free(g_device_list[nb]);
			break;
			};

		// Okay, let's add a new device to global list
		strcpy(g_device_name_list[nb], name);

		dev = g_device_list[nb];
			
		memset(dev, 0, sizeof(*dev));

		dev->id	  			= nb;
		dev->name 			= g_device_name_list[nb];
		dev->flags			= 0;
		dev->open_count		= 0;

		dev->data			= g_fake;	// opaque *thing* !!!
				
		dev->init_semaphore	= create_sem(0, DRIVER_NAME "_init_semaphore");	// 0 -> block on first acquire
		set_sem_owner(dev->init_semaphore, B_SYSTEM_TEAM); 	// Make kernel owner of init_semaphore

		DEBUG_ONLY( dprintf(DRIVER_NAME ": -> %s\n", dev->name) );
		};

	// Mark end of list
	g_device_list[nb] = NULL;
	g_device_name_list[nb] = NULL;

	return ( nb > 0) ? B_OK : ENODEV ;
}


// -------------------------------------------------------
_EXPORT void uninit_driver(void)
{
	long 	i; 

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- uninit_driver()\n") );

	// Free device list
	for ( i = 0; g_device_list[i]; i++ )
		{
		device * dev;
		
		dev = g_device_list[i];
		if ( dev->init_semaphore >= 0 )
			delete_sem(dev->init_semaphore);
		
		free(dev);
		};

	// Free device name list
	for ( i = 0; g_device_name_list[i]; i++ )
		free(g_device_name_list[i]);
}


// -------------------------------------------------------
_EXPORT const char ** publish_devices(void)
{
#ifdef DEBUG
	{
	long 	i;
	
	// Display published devices names
	dprintf(DRIVER_NAME ": ---- publish_devices():\n");

	for ( i = 0; g_device_name_list[i]; i++ )
		dprintf(DRIVER_NAME ": %s\n", g_device_name_list[i] );
	};
#endif

	return (const char **) g_device_name_list;
}


// -------------------------------------------------------
_EXPORT device_hooks * find_device
	(
	const char *	name
	)
{
	long 	i;
	char *	device_name;

	// Find device name
	for ( i = 0; (device_name = g_device_name_list[i]); i++ )
		{
		if ( strcmp(device_name, name) == 0 )
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- find_device: %s published, great !\n", name) );
			return &g_device_hooks;
			};
		};

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- find_device: %s NOT PUBLISHED, ARGH !\n", name) );
	return NULL; // We did'nt published this device name, it's weird, what's up, Kernel ?!?
}


#ifdef CODEWARRIOR
	#pragma mark [Device hooks]
#endif

// Device hooks
// ------------

// -------------------------------------------------------
static status_t open_hook
	(
	const char *	name, 
	uint32 			flags, 
	void **			cookie
	)
{
	device *	dev;
	long		i;
	status_t	status;

	// Find device name
	for ( i = 0; (dev = g_device_list[i]); i++ )
		{
		if ( strcmp(dev->name, name) == 0 )
			break;
		};

	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: %s NOT FOUND, ARGH !\n", name) );
		return EINVAL;
		};
		
	if ( atomic_add(&(dev->open_count), 1) == 0 )
		{
		// First open on this device :
		status = device_init(dev);
		if ( status != B_OK )
			{
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: device_init(%s)->%d, FAILURE !\n", dev->name, status) );
			return status;
			};
		}
	else
		// Not the first open on this device
		// But perhaps the first 'open' on this device is still pending,
		// so we wait for end (death !) of init_semaphore
		acquire_sem(dev->init_semaphore);

	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- open_hook: %s opened, great !\n", dev->name) );

	*cookie = dev;

	return B_OK; // B_ERROR;
}


// -------------------------------------------------------
static status_t close_hook
	(
	void *	cookie
	)
{
	device *	dev;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- close_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- close_hook %s\n", dev->name) );

	return B_OK;
}


// -------------------------------------------------------
static status_t free_hook
	(
	void *	cookie
	)
{
	device *	dev;
	status_t	status;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- free_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- free_hook %s\n", dev->name) );

	if ( atomic_add(&(dev->open_count), -1) == 1 )
		// Last free() on this device :
		status = device_uninit(dev);
	else
		// not the last session, nothing to actually free
		status = B_OK;
	
	return status;
}



// -------------------------------------------------------
static status_t control_hook
	(
	void *	cookie, 
	uint32 	op_code, 
	void *	data, 
	size_t 	len
	)
{
	device *	dev;
	FAKE *		f;

	dev = (device *) cookie;
	if ( (NULL == dev) || (dev->data == NULL) )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
		
	f = dev->data;
	
	switch ( op_code ) {
	case SENSOR_GET_MONITOR_INFO:
		{
		SENSORS_MONITOR	* 	sm;
		
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_GET_MONITOR_INFO\n", dev->name) );
		
		// Hum, seem that 'len' is ALWAYS zero !!!
		// if ( len < sizeof(*sm) )
		// 	 return EINVAL;
		
		sm = (SENSORS_MONITOR *) data;
		
		sm->api_version = SENSOR_CURRENT_API_VERSION;
		strncpy(sm->name, f->name, sizeof(sm->name));
		sm->nb_sensors = dev->nb_sensors;
		break;
		};

	case SENSOR_GET_VALUE:
		{
		SENSOR *	s;
		int			id;
		
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_GET_VALUE\n", dev->name) );

		// Hum, seem that 'len' is ALWAYS zero !!!
		// if ( len < sizeof(*s) )
		//   return EINVAL;
			
		s = (SENSOR *) data;
		id = s->id;
		if ( id < 0 || id > (dev->nb_sensors - 1) )
			return EINVAL;
			
		fake_update_values(dev);
			
		memcpy(s, &(dev->sensors[id]), sizeof(*s));
		s->when = system_time();
		break;
		};

	case SENSOR_COMMAND:
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: SENSOR_COMMAND\n", dev->name) );
		return EINVAL;
		};

	default:
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- control_hook %s: op_code %d (%x) unhandled.\n", dev->name, op_code, op_code) );
		return EINVAL;
		};
	}
	return B_OK;
}


// -------------------------------------------------------
static status_t read_hook
	(
	void *		cookie, 
	off_t 		pos, 
	void *		data, 
	size_t *	len
	)
{
	device *	dev;
	int			i;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- read_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- read_hook %s: %x, %d\n",
				dev->name, data, *len ) );
	
	/*
	for( i = 0; i < dev->nb_sensors; i++ )
		{
		sprintf( data, "%d\t[%s]\t%f\n",
				 dev->sensors[i].id, dev->sensors[i].name, dev->sensors[i].value ); 
		};
*/
	// *len = strlen(data);
	// return B_OK;

	*len = 0;
	return B_IO_ERROR;
}



// -------------------------------------------------------
static status_t write_hook
	(
	void *			cookie, 
	off_t 			pos, 
	const void *	data, 
	size_t *		len
	)
{
	device *	dev;

	dev = (device *) cookie;
	if ( NULL == dev )
		{
		DEBUG_ONLY( dprintf(DRIVER_NAME ": --- write_hook COOKIE BURNED !\n") );
		return B_ERROR;
		};
	
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- write_hook %s: %x, %d\n",
				dev->name, data, *len ) );

	*len = 0;
	return B_IO_ERROR;
}


#ifdef CODEWARRIOR
	#pragma mark [Privates routines]
#endif

// Privates routines
// -----------------


// -------------------------------------------------------
static status_t device_init
	(
	device *	dev
	)
{
	uint32			size;
	int 			i;
	status_t		status;

	status = B_ERROR;
	
	if ( NULL == dev )
		{
		status = B_BAD_VALUE;
		goto EXIT;
		};
		
	DEBUG_ONLY( dprintf(DRIVER_NAME ": --- device_init %s\n", dev->name) );
		
	dev->update_semaphore	= create_sem(1, DRIVER_NAME "_update_semaphore");
	set_sem_owner(dev->update_semaphore, B_SYSTEM_TEAM); // Make kernel owner of update_semaphore

	dev->last_updated		= 0;	// not yet updated values
	
	status = fake_init(dev);

EXIT:;

	delete_sem(dev->init_semaphore); // kill / release concurrent device_init()
	dev->init_semaphore = -1;
	
	return status;
}


// -------------------------------------------------------
static status_t device_uninit
	(
	device *	dev
	)
{
	fake_uninit(dev);

	delete_sem(dev->update_semaphore);
	dev->update_semaphore = -1;

	return B_OK;
}


#ifdef CODEWARRIOR
	#pragma mark [Tools]
#endif


// -------------------------------------------------------
static SENSOR *	device_find_sensor
	(
	device *	dev,	
	char *		name,
	int			id
	)
{
	int			i;

	if ( NULL == dev )
		return NULL;
		
	if ( NULL == dev->sensors )
		return NULL;

	for( i = 0; i < dev->nb_sensors; i++ )
		{
		if ( (name != NULL) && (strcmp(name, dev->sensors[i].name) == 0) )
			return &(dev->sensors[i]);
			
		if ( (name == NULL) && (dev->sensors[i].id == id) )
			return &(dev->sensors[i]);
		};

	return NULL;		
}




// -------------------------------------------------------
static status_t fake_init
	(
	device *	dev
	)
{
	FAKE * 			f;
	FAKE_SENSORS * 	p;
	
	if ( (NULL == dev) || (dev->data == NULL) )
		return B_BAD_VALUE;

	// dprintf(DRIVER_NAME ": w8378xx_init: sensors list:\n");
	
	f = dev->data;

	// How many sensors ?
	dev->nb_sensors 	= 0;
	for ( p = g_fake_sensors; p->name != NULL; p++ )
			dev->nb_sensors++;

	// Make room for sensors datas
	dev->sensors		= (SENSOR *) malloc((dev->nb_sensors * sizeof(SENSOR)) + 16);	

	// Load sensors descriptions & stuff
	dev->nb_sensors 	= 0;
	for ( p = g_fake_sensors; p->name != NULL; p++ )
		{
		SENSOR * s;

		s = &(dev->sensors[dev->nb_sensors]);
			
		s->id 				= dev->nb_sensors;

		strncpy(s->name, p->name, sizeof(s->name));
		s->kind 			= p->kind;
		s->flags 			= p->flags;
		s->when				= 0;	// need to be updated !!!
		s->value 			= 0;
		s->value_multiplier = p->value_multiplier;
		s->value_divider	= p->value_divider;
				
		// dprintf(DRIVER_NAME ": %d\t%s\t%d\t%d\n",
		//	    	s->id, s->name, s->kind, s->value );
						
		dev->nb_sensors++;
		};

	return B_OK;
}


// -------------------------------------------------------
static status_t fake_uninit
	(
	device * dev
	)
{
	if ( NULL == dev )
		return B_BAD_VALUE;

	if ( dev->sensors != NULL )
		free(dev->sensors);
	dev->sensors = NULL;

	return B_OK;
}


// -------------------------------------------------------
static status_t fake_update_values
	(
	device * dev
	)
{
	status_t	status;
	FAKE *		f;

	if ( NULL == dev )
		return B_BAD_VALUE;

	f = dev->data;
	
	// Okay, acquire update sem right now, or just wait the pending one finish update phase
	status = acquire_sem_etc( dev->update_semaphore, 1, B_RELATIVE_TIMEOUT, 0 );
	if( status == B_NO_ERROR )
		{
		bigtime_t	when;
		
		when = system_time();
		
		if( when - dev->last_updated > FAKE_UPDATE_PERIOD )
			{
			SENSOR *	s;
			
			// Need a update
			DEBUG_ONLY( dprintf(DRIVER_NAME ": --- fake_update_values\n") );
		
			// read registers values
			dev->last_updated = when;
			
			// Okay, values updating -> read the registers values

			// ---- VCore A (2V) ----
			if ( (s = device_find_sensor(dev, "in0", 0)) != NULL )
				s->value = 2200;
				
			// ---- VCore B (2V) ----
			if ( (s = device_find_sensor(dev, "in1", 0)) != NULL )
				s->value = 2200;
				
			// ---- 3.3V ----
			if ( (s = device_find_sensor(dev, "in2", 0)) != NULL )
				s->value = 3300;
				
			// ---- 5V ----
			if ( (s = device_find_sensor(dev, "in3", 0)) != NULL )
				s->value = 5500;
				
			// ---- 12V ----
			if ( (s = device_find_sensor(dev, "in4", 0)) != NULL )
				s->value = 12000;
				
			// ---- -12V ----
			if ( (s = device_find_sensor(dev, "in5", 0)) != NULL )
				s->value = -12000;
				
			// ---- -5V ----
			if ( (s = device_find_sensor(dev, "in6", 0)) != NULL )
				s->value = -5000;
				
			// ---- Hum? ----
			if ( (s = device_find_sensor(dev, "in7", 0)) != NULL )
				s->value = 0;
				
			// ---- Hum?, but bis ----
			if ( (s = device_find_sensor(dev, "in8", 0)) != NULL )
				s->value = 0;
				
			// ---- Temperature 1 (motherboard, often) ----
			if ( (s = device_find_sensor(dev, "temp1", 0)) != NULL )
				s->value = 270;
				
			// ---- Temperature 2 (CPU, very often) ----
			if ( (s = device_find_sensor(dev, "temp2", 0)) != NULL )
				s->value = 345;
				
			// ---- Temperature 3 (second CPU or other stuffs, like disks) ----
			if ( (s = device_find_sensor(dev, "temp3", 0)) != NULL )
				s->value = 352;
				
			// ---- Fan 1 ----
			if ( (s = device_find_sensor(dev, "fan1", 0)) != NULL )
				s->value = 2470;
				
			// ---- Fan 2 ----
			if ( (s = device_find_sensor(dev, "fan2", 0)) != NULL )
				s->value = 5120;
				
			// ---- Fan 3 ----
			if ( (s = device_find_sensor(dev, "fan3", 0)) != NULL )
				s->value = 4970;

			// ---- Voltage ID ----
			if ( (s = device_find_sensor(dev, "vid", 0)) != NULL )
				s->value = 1800;

			dev->nb_updates++;
			// Okay, it's done.
			};
		}
	else if ( status = B_WOULD_BLOCK )
		// okay, concurrent thread run the update process, just wait it's end
		acquire_sem(dev->update_semaphore);
	else
		// 	Argh, something going wrong with update semaphore !
		return status;
		 
	// Update is done, by this thread or a concurrent thread, but who mind ? 
	release_sem( dev->update_semaphore );
	
	return B_OK;
}



